<?php

namespace Drupal\choices_autocomplete\Plugin\Field\FieldWidget;

use Drupal\Component\Serialization\Json;
use Drupal\Component\Utility\Xss;
use Drupal\Core\Entity\Element\EntityAutocomplete;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityReferenceSelection\SelectionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Field\Plugin\Field\FieldWidget\EntityReferenceAutocompleteWidget;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element\FormElement;
use Drupal\Core\Render\ElementInfoManagerInterface;
use Drupal\views\Entity\View;
use Drupal\views\Plugin\EntityReferenceSelection\ViewsSelection;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Validator\ConstraintViolationInterface;

/**
 * Plugin implementation of the 'entity_reference_choices' widget.
 *
 * @FieldWidget(
 *   id = "entity_reference_choices",
 *   label = @Translation("Choices.js autocomplete"),
 *   description = @Translation("An autocomplete text field powered by Choices.js."),
 *   field_types = {
 *     "entity_reference"
 *   },
 *   multiple_values = TRUE
 * )
 */
class EntityReferenceChoicesWidget extends EntityReferenceAutocompleteWidget {

  /**
   * The element info manager.
   *
   * @var \Drupal\Core\Render\ElementInfoManagerInterface
   */
  protected ElementInfoManagerInterface $elementInfoManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->elementInfoManager = $container->get('plugin.manager.element_info');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings(): array {
    return [
      'no_choices_text' => t('Start typing to find results'),
      'loading_text' => t('Loading...'),
      'no_results_text' => t('No results found'),
      'item_select_text' => t('Press to select'),
      'max_item_text' => t('Max number of items selected'),
      'input_pattern' => NULL,
      'input_minlength' => 0,
      'input_maxlength' => 64,
      'position' => 'auto',
    ] + parent::defaultSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state): array {
    $element = parent::settingsForm($form, $form_state);
    unset($element['size']);

    $element['no_choices_text'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Start typing text'),
      '#description' => $this->t('Text shown in dropdown until user search input is detected.'),
      '#default_value' => $this->getSetting('no_choices_text'),
    ];
    $element['loading_text'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Searching text'),
      '#description' => $this->t('Text shown in dropdown as search results are being retrieved.'),
      '#default_value' => $this->getSetting('loading_text'),
    ];
    $element['no_results_text'] = [
      '#type' => 'textfield',
      '#title' => $this->t('No results text'),
      '#description' => $this->t('Text shown in dropdown when no results were found.'),
      '#default_value' => $this->getSetting('no_results_text'),
    ];
    $element['item_select_text'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Press to select text'),
      '#description' => $this->t('Text shown in dropdown next to the current highlighted result.'),
      '#default_value' => $this->getSetting('item_select_text'),
    ];
    $element['max_item_text'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Max items text'),
      '#description' => $this->t('Text shown in dropdown when the maximum number of items are selected.'),
      '#default_value' => $this->getSetting('max_item_text'),
    ];
    $element['input_pattern'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Allowed input pattern'),
      '#description' => $this->t('Use regexp pattern to limit the allowed user input search characters.'),
      '#default_value' => $this->getSetting('input_pattern'),
    ];
    $element['input_minlength'] = [
      '#type' => 'number',
      '#title' => $this->t('Search input minimum length'),
      '#description' => $this->t('Search when user input is equal to or exceeds a minimum number of characters.'),
      '#min' => 0,
      '#default_value' => $this->getSetting('input_minlength'),
    ];
    $element['input_maxlength'] = [
      '#type' => 'number',
      '#title' => $this->t('Search input maximum length'),
      '#description' => $this->t('Limit user search input to a maximum number of characters.'),
      '#min' => 0,
      '#default_value' => $this->getSetting('input_maxlength'),
    ];
    $element['position'] = [
      '#type' => 'select',
      '#title' => $this->t('Dropdown position'),
      '#description' => $this->t('Change the search results dropdown position.'),
      '#options' => [
        'auto' => $this->t('Automatic (recommended)'),
        'top' => $this->t('Top'),
        'bottom' => $this->t('Bottom'),
      ],
      '#default_value' => $this->getSetting('position'),
    ];

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary(): array {
    $summary = parent::settingsSummary();

    if ($value = $this->getSetting('no_choices_text')) {
      $summary[] = $this->t('Start typing text: @value', ['@value' => $value]);
    }
    if ($value = $this->getSetting('loading_text')) {
      $summary[] = $this->t('Searching text: @value', ['@value' => $value]);
    }
    if ($value = $this->getSetting('no_results_text')) {
      $summary[] = $this->t('No results text: @value', ['@value' => $value]);
    }
    if ($value = $this->getSetting('item_select_text')) {
      $summary[] = $this->t('Item select text: @value', ['@value' => $value]);
    }
    if ($value = $this->getSetting('max_item_text')) {
      $summary[] = $this->t('Max items text: @value', ['@value' => $value]);
    }
    if ($value = $this->getSetting('input_pattern')) {
      $summary[] = $this->t('Allowed input pattern: @value', ['@value' => $value]);
    }
    if ($value = $this->getSetting('input_minlength')) {
      $summary[] = $this->t('Input minimum length: @value', ['@value' => $value]);
    }
    if ($value = $this->getSetting('input_maxlength')) {
      $summary[] = $this->t('Input maximum length: @value', ['@value' => $value]);
    }
    if ($value = $this->getSetting('position')) {
      $summary[] = $this->t('Dropdown position: @value', ['@value' => $value]);
    }

    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state): array {
    /** @var \Drupal\Core\Field\EntityReferenceFieldItemListInterface $items */
    $cardinality = $items->getFieldDefinition()->getFieldStorageDefinition()->getCardinality();

    // Use select element as base instead of text input. Choices.js binds
    // to the select and adds all the necessary markup and text input.
    $options = [];
    if ($placeholder = $this->getSetting('placeholder')) {
      $options[''] = $placeholder;
    }
    $element += [
      '#type' => 'select',
      '#options' => $options,
      '#multiple' => $cardinality === FieldStorageDefinitionInterface::CARDINALITY_UNLIMITED || $cardinality > 1,
      '#cardinality' => $cardinality,
      '#process' => array_merge(
        [
          [EntityAutocomplete::class, 'processEntityAutocomplete'],
          [FormElement::class, 'processAutocomplete'],
          [static::class, 'processEntityAutocomplete'],
        ],
        $this->elementInfoManager->getInfo('select')['#process'],
      ),
      '#element_validate' => [[static::class, 'validateEntityAutocomplete']],
      '#no_choices' => TRUE,
      '#attributes' => [
        'class' => ['js-choices-autocomplete'],
        'data-choices' => Json::encode([
          'maxItemCount' => $cardinality,
          'searchResultLimit' => $this->getSetting('match_limit'),
          'noChoicesText' => Xss::filterAdmin($this->getSetting('no_choices_text')),
          'loadingText' => Xss::filterAdmin($this->getSetting('loading_text')),
          'noResultsText' => Xss::filterAdmin($this->getSetting('no_results_text')),
          'itemSelectText' => Xss::filterAdmin($this->getSetting('item_select_text')),
          'maxItemText' => Xss::filterAdmin($this->getSetting('max_item_text')),
          'position' => $this->getSetting('position'),
        ]),
        'data-placeholder' => $placeholder,
      ],
      '#default_value' => [],
    ];

    // Set default values and additional options.
    foreach ($items->referencedEntities() as $entity) {
      $key = $entity->isNew() ? $entity->label() : "{$entity->label()} ({$entity->id()})";
      $element['#default_value'][$key] = $key;
    }
    if ($pattern = strip_tags($this->getSetting('input_pattern'))) {
      $element['#attributes']['data-pattern'] = $pattern;
    }
    if ($minlength = $this->getSetting('input_minlength')) {
      $element['#attributes']['data-minlength'] = $minlength;
    }
    if ($maxlength = $this->getSetting('input_maxlength')) {
      $element['#attributes']['data-maxlength'] = $maxlength;
    }

    $element += parent::formElement($items, $delta, $element, $form, $form_state)['target_id'];
    $element += ['#autocreate' => FALSE];
    if ($element['#autocreate']) {
      $element['#autocreate']['pattern'] = $pattern;
      $element['#autocreate']['minlength'] = $minlength;
      $element['#autocreate']['maxlength'] = $maxlength;
      $element['#attributes']['data-autocreate'] = 'true';
    }
    unset($element['#maxlength'], $element['#size']);
    $element['#attached']['library'][] = 'choices_autocomplete/choices';

    return $element;
  }

  /**
   * Process entity autocomplete element.
   *
   * Extracts and attempts to format labels for the values. Labels are passed
   * in drupalSettings because option elements cannot include HTML. Values
   * with formatted labels are replaced when Choices.js initializes.
   */
  public static function processEntityAutocomplete(array &$element, FormStateInterface $form_state, array &$complete_form): array {
    if ($values = $element['#value'] ?: $element['#default_value']) {
      $values = is_array($values) ? $values : [$values];
      $labels = [];
      foreach ($values as $value) {
        if ($entity = static::extractEntityFromInput($value, $element)) {
          $element['#options'][$value] = $value;

          if (!$entity->isNew()) {
            $labels["{$entity->label()} ({$entity->id()})"] = $entity->entity_autocomplete_label;
          }
        }
      }
      $element['#attached']['drupalSettings']['choices_autocomplete'] = $labels;
    }
    return $element;
  }

  /**
   * Validate entity autocomplete element.
   */
  public static function validateEntityAutocomplete(array &$element, FormStateInterface $form_state, array &$complete_form): void {
    $values = array_filter(is_array($element['#value']) ? $element['#value'] : [$element['#value']]);
    $newValues = [];
    foreach ($values as $value) {
      if ($entity = static::extractEntityFromInput($value, $element)) {
        if ($entity->isNew()) {
          $value = preg_replace('/\s\(\d+\)$/', '', $value);
          if (isset($element['#autocreate']['pattern']) && ($chars = static::validateInputPattern($value, $element['#autocreate']['pattern']))) {
            $form_state->setError($element, t('Invalid characters: %value', [
              '%value' => implode('', $chars),
            ]));
          }
          elseif (isset($element['#autocreate']['minlength']) && mb_strlen($value) < $element['#autocreate']['minlength']) {
            $form_state->setError($element, t('Too few characters: %value', [
              '%value' => $value,
            ]));
          }
          elseif (isset($element['#autocreate']['maxlength']) && mb_strlen($value) > $element['#autocreate']['maxlength']) {
            $form_state->setError($element, t('Too many characters: %value', [
              '%value' => $value,
            ]));
          }
          else {
            $newValues[] = ['entity' => $entity];
          }
        }
        else {
          $newValues[] = ['target_id' => $entity->id()];
        }
      }
      else {
        $form_state->setError($element, t('Invalid selections.'));
      }
    }
    $form_state->setValueForElement($element, $newValues);
  }

  /**
   * Validate input pattern for auto-created items.
   *
   * @param string $value
   *   The value.
   * @param string $pattern
   *   The regular expression pattern.
   *
   * @return array
   *   Returns an array of invalid characters.
   */
  public static function validateInputPattern(string $value, string $pattern): array {
    $chars = [];
    foreach (mb_str_split($value) as $char) {
      if (!preg_match("/{$pattern}/", $char)) {
        $chars[$char] = $char;
      }
    }
    return $chars;
  }

  /**
   * {@inheritdoc}
   */
  public function errorElement(array $element, ConstraintViolationInterface $error, array $form, FormStateInterface $form_state) {
    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function massageFormValues($values, array $form, FormStateInterface $form_state): array {
    return $values;
  }

  /**
   * Extract entity from user input.
   *
   * @param string $input
   *   The user input.
   * @param array $element
   *   The element.
   *
   * @return \Drupal\Core\Entity\EntityInterface
   *   Returns the fully loaded extracted entity.
   */
  public static function extractEntityFromInput(string $input, array $element): ?EntityInterface {
    $storage = \Drupal::entityTypeManager()->getStorage($element['#target_type']);
    $handler = static::getSelectionHandler($element);

    if ($id = (int) EntityAutocomplete::extractEntityIdFromAutocompleteInput($input)) {
      if ($entity = $storage->load($id)) {
        $matches = [];

        // Views selection handlers are extra fancy in that they can
        // match against any property or field. To extract and format a label
        // the underlying views display must be executed. Passing the
        // extracted ID ensures the view only includes that entity.
        if ($handler instanceof ViewsSelection) {
          $config = $handler->getConfiguration();

          $view = View::load($config['view']['view_name']);
          $view = $view->getExecutable();
          $view->setDisplay($config['view']['display_name']);
          $view->setArguments($config['view']['arguments']);
          $view->display_handler->setOption('entity_reference_options', [
            'limit' => 0,
            'ids' => [$id],
          ]);
          $view->execute();
          $rendered = $view->render();

          if (isset($rendered[$entity->id()])) {
            $matches[$entity->bundle()][$entity->id()] = \Drupal::service('renderer')->renderRoot($rendered[$entity->id()]);
          }
        }
        // Use default behavior for all other selection handlers.
        else {
          $matches = $handler->getReferenceableEntities($entity->label(), '=', 100);
        }

        if (isset($matches[$entity->bundle()][$entity->id()])) {
          $entity = clone $entity;
          $entity->entity_autocomplete_label = $matches[$entity->bundle()][$entity->id()];
          return $entity;
        }
      }
    }

    // Auto-create new entity with the user input.
    elseif ($element['#autocreate']) {
      /** @var \Drupal\Core\Entity\EntityReferenceSelection\SelectionWithAutocreateInterface $handler */
      return $handler->createNewEntity($element['#target_type'], $element['#autocreate']['bundle'], $input, $element['#autocreate']['uid']);
    }

    return NULL;
  }

  /**
   * Get entity autocomplete selection handler.
   *
   * @param array $element
   *   The element.
   *
   * @return \Drupal\Core\Entity\EntityReferenceSelection\SelectionInterface
   *   Returns the selection handler plugin instance.
   */
  public static function getSelectionHandler(array $element): SelectionInterface {
    return \Drupal::service('plugin.manager.entity_reference_selection')
      ->getInstance($element['#selection_settings'] + [
        'target_type' => $element['#target_type'],
        'handler' => $element['#selection_handler'],
      ]);
  }

}
